(* interface to UFTP client(s) *)
(* solyga@gmx.de, 2021-01-13 *)
IMPLEMENTATION MODULE CliIO;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT WinSock;
IMPORT Lib2, Lib3, SO, Utils, Input, Config, Tmr, Tmrs, Cfg, Log;

CONST
  mid = Cfg.midCliIO;
  period = 500;  (* s *)

TYPE
  TSocket  = WinSock.SOCKET;
  TSockSet = WinSock.fd_set;
  TAdrDesc = WinSock.sockaddr_in;

CONST
  adrNull = 0;  (* = WinSock.INADDR_ANY *)
  prtNull = 0;  (* yields ephemeral port *)
  socNull = WinSock.INVALID_SOCKET;  (* -1 *)

TYPE
  TCntDesc = RECORD
    txn,
    txe: CARDINAL;
    rxn,
    rxe: CARDINAL;
  END;

VAR
  hnd: THandler;
  soc: TSocket;
  cnt: TCntDesc;
  tcs: Tmr.TTicks;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, Config.data.vrb );
  HALT;  (* don't remove this before having checked return code implications *)
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE Trigger();

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..31] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Trigger ' );
    SO.WrLngHex( b, tcs, 8 );
    Message( s );
  END Show;

VAR
  now: Tmr.TTicks;
BEGIN
  Input.Trigger();
  now := Tmr.Ticks();
  Tmrs.Trigger( Tmr.Diff( tcs, now ) );
  tcs := now;
  Show();
END Trigger;


PROCEDURE NewSock(): TSocket;
VAR
  s: TSocket;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  s := WinSock.socket( WinSock.AF_INET, WinSock.SOCK_DGRAM, WinSock.IPPROTO_UDP );
  IF s = WinSock.INVALID_SOCKET THEN
    SO.Init( b, t );
    SO.WrStr( b, 'NewSock socket() failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( t );
  END;
  RETURN s;
END NewSock;

PROCEDURE AdjSock( s: TSocket );

  PROCEDURE Show( s-: ARRAY OF CHAR; l: CARDINAL );
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    t: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, t );
    SO.WrStr( b, 'AdjSock ' );
    SO.WrStr( b, s );
    SO.WrStr( b, ' ' );
    SO.WrLngCard( b, l, 1 );
    Message( t );
  END Show;

  PROCEDURE SetNonBlock( s: TSocket );  (* no get possible :-/ *)
  VAR
    x: CARDINAL;
    b: SO.TBuffer;
    t: ARRAY [0..63] OF CHAR;
  BEGIN
    x := 1;  (* non-blocking *)
    IF WinSock.ioctlsocket( s, WinSock.FIONBIO, ADR(x) ) # 0 THEN
      SO.Init( b, t );
      SO.WrStr( b, 'SetNonBlock ioctlsocket() failed, err ' );
      SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
      Fatal( t );
    END;
    Show( 'SetNonBlock', x );
  END SetNonBlock;

  PROCEDURE RcvLen( s: TSocket ): CARDINAL;
  VAR
    x, l, r: INTEGER;
    b: SO.TBuffer;
    t: ARRAY [0..63] OF CHAR;
  BEGIN
    x := 0; l := SIZE( x );
    r := WinSock.getsockopt( s, WinSock.SOL_SOCKET, WinSock.SO_RCVBUF, ADR(x), l );
    IF r = WinSock.SOCKET_ERROR THEN
      SO.Init( b, t );
      SO.WrStr( b, 'RcvLen getsockopt() failed, err ' );
      SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
      Fatal( t );
    END;
    IF l # 4 THEN
      SO.Init( b, t );
      SO.WrStr( b, 'RcvLen getsockopt() returned invalid par len ' );
      SO.WrLngInt( b, l, 1 );
      Fatal( t );
    END;
    Show( 'RcvLen', CARDINAL(x) );
    RETURN CARDINAL( x );
  END RcvLen;

  PROCEDURE SetRcvLen( s: TSocket; v: CARDINAL );
  VAR
    x, r: INTEGER;
    b: SO.TBuffer;
    t: ARRAY [0..63] OF CHAR;
  BEGIN
    x := INTEGER( v );
    r := WinSock.setsockopt( s, WinSock.SOL_SOCKET, WinSock.SO_RCVBUF, ADR(x), SIZE(x) );
    IF r = WinSock.SOCKET_ERROR THEN
      SO.Init( b, t );
      SO.WrStr( b, 'SetRcvLen setsockopt() failed, err ' );
      SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
      Fatal( t );
    END;
    Show( 'SetRcvLen', v );
  END SetRcvLen;

  PROCEDURE SndLen( s: TSocket ): CARDINAL;
  VAR
    x, l, r: INTEGER;
    b: SO.TBuffer;
    t: ARRAY [0..63] OF CHAR;
  BEGIN
    x := 0; l := SIZE( x );
    r := WinSock.getsockopt( s, WinSock.SOL_SOCKET, WinSock.SO_SNDBUF, ADR(x), l );
    IF r = WinSock.SOCKET_ERROR THEN
      SO.Init( b, t );
      SO.WrStr( b, 'SndLen getsockopt() failed, err ' );
      SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
      Fatal( t );
    END;
    IF l # 4 THEN
      SO.Init( b, t );
      SO.WrStr( b, 'SndLen getsockopt() returned invalid par len ' );
      SO.WrLngInt( b, l, 1 );
      Fatal( t );
    END;
    Show( 'SndLen', CARDINAL(x) );
    RETURN CARDINAL( x );
  END SndLen;

  PROCEDURE SetSndLen( s: TSocket; v: CARDINAL );
  VAR
    x, r: INTEGER;
    b: SO.TBuffer;
    t: ARRAY [0..63] OF CHAR;
  BEGIN
    x := INTEGER( v );
    r := WinSock.setsockopt( s, WinSock.SOL_SOCKET, WinSock.SO_SNDBUF, ADR(x), SIZE(x) );
    IF r = WinSock.SOCKET_ERROR THEN
      SO.Init( b, t );
      SO.WrStr( b, 'SetSndLen setsockopt() failed, err ' );
      SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
      Fatal( t );
    END;
    Show( 'SetSndLen', v );
  END SetSndLen;

  PROCEDURE MsgLen( s: TSocket ): CARDINAL;
  VAR
    x: CARDINAL;
    l, r: INTEGER;
    b: SO.TBuffer;
    t: ARRAY [0..63] OF CHAR;
  BEGIN
    x := 0; l := SIZE( x );
    r := WinSock.getsockopt( s, WinSock.SOL_SOCKET, WinSock.SO_MAX_MSG_SIZE, ADR(x), l );
    IF r = WinSock.SOCKET_ERROR THEN
      SO.Init( b, t );
      SO.WrStr( b, 'MsgLen getsockopt() failed, err ' );
      SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
      Fatal( t );
    END;
    IF l # 4 THEN
      SO.Init( b, t );
      SO.WrStr( b, 'MsgLen getsockopt() returned invalid par len ' );
      SO.WrLngInt( b, l, 1 );
      Fatal( t );
    END;
    Show( 'MsgLen', x );
    RETURN x;
  END MsgLen;

CONST
  (* length = bandwidth * time *)
  lenRcv = 128 * 1024;  (* 128kB, 2ms buffering => 62MB/s *)
  lenSnd = 128 * 1024;
VAR
  val: CARDINAL;
BEGIN
  SetNonBlock( s );

  val := RcvLen( s );
  IF val < lenRcv THEN
    SetRcvLen( s, lenRcv );
    val := RcvLen( s );
  END;

  val := SndLen( s );
  IF val < lenSnd THEN
    SetSndLen( s, lenSnd );
    val := SndLen( s );
  END;

  val := MsgLen( s );
END AdjSock;

PROCEDURE BindSock( s: TSocket; la: TAddress; lp: TPort );
VAR
  a: TAdrDesc;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  a.sin_family      := WinSock.AF_INET;
  a.sin_addr.s_addr := la;
  a.sin_port        := WinSock.htons( lp );
  IF WinSock.bind( s, ADR(a), SIZE(a) ) = WinSock.SOCKET_ERROR THEN
    SO.Init( b, t );
    SO.WrStr( b, 'BindSock bind() for ' );
    Utils.WrIP( b, la );
    SO.WrChar( b, ':' );
    SO.WrCard( b, lp, 1 );
    SO.WrStr( b, ' failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( t );
  END;
END BindSock;

PROCEDURE GetSockAdr( s: TSocket; VAR la: TAddress; VAR lp: TPort );
VAR
  a: TAdrDesc;
  l: INTEGER;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  Lib2.Fill( a, 0 );
  l := SIZE( a );
  IF WinSock.getsockname( s, ADR(a), l ) = WinSock.SOCKET_ERROR THEN
    SO.Init( b, t );
    SO.WrStr( b, 'GetSockAdr getsockname() failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( t );
  END;
  IF a.sin_family # WinSock.AF_INET THEN
    SO.Init( b, t );
    SO.WrStr( b, 'GetSockAdr getsockname() returned invalid family value' );
    Fatal( t );
  END;
  IF l >= 8 THEN
    la := a.sin_addr.s_addr;
    lp := WinSock.ntohs( a.sin_port );
  ELSE
    SO.Init( b, t );
    SO.WrStr( b, 'GetSockAdr getsockname() returned short address record' );
    Fatal( t );
  END;
END GetSockAdr;

PROCEDURE WrSock( s: TSocket; ra: TAddress; rp: TPort; x-: ARRAY OF BYTE; l: CARDINAL );
VAR
  a: TAdrDesc;
  r: INTEGER;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  Lib2.Fill( a, 0 );
  a.sin_family      := WinSock.AF_INET;
  a.sin_port        := WinSock.htons( rp );
  a.sin_addr.s_addr := ra;
  r := WinSock.sendto( s, ADR(x), INTEGER(l), 0, ADR(a), SIZE(a) );
  IF r = WinSock.SOCKET_ERROR THEN
    SO.Init( b, t );
    SO.WrStr( b, 'WrSock sendto() failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( t );
  END;
  IF r # INTEGER( l ) THEN
    SO.Init( b, t );
    SO.WrStr( b, 'WrSock sendto() failed, ' );
    SO.WrLngInt( b, r, 1 );
    SO.WrStr( b, ' of ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, ' bytes written' );
    Fatal( t );
  END;
END WrSock;

PROCEDURE RdSock( s: TSocket; VAR ra: TAddress; VAR rp: TPort; VAR x: ARRAY OF BYTE ): CARDINAL;
VAR
  a: TAdrDesc;
  al: INTEGER;
  r: INTEGER;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  Lib2.Fill( a, 0 );
  al := SIZE( a );
  r := WinSock.recvfrom( s, ADR(x), SIZE(x), 0, ADR(a), al );
  (* buffer too small  =>  r = -1, WSAEMSGSIZE (10040), excess data lost *)
  (* nonblocking & no data available  =>  r = -1, WSAEWOULDBLOCK (10035) *)
  (* further possible errors: WSAECONNRESET (10054) *)
  IF r < 0 THEN
    SO.Init( b, t );
    SO.WrStr( b, 'RdSock recvfrom() failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( t );
  END;
  IF al < 8 THEN
    SO.Init( b, t );
    SO.WrStr( b, 'RdSock recvfrom() failed, al ' );
    SO.WrLngInt( b, al, 1 );
    SO.WrStr( b, ' < 8' );
    Fatal( t );
  END;
  IF CARDINAL( r ) > SIZE( x ) THEN
    SO.Init( b, t );
    SO.WrStr( b, 'RdSock recvfrom() failed, implausible return value ' );
    SO.WrLngInt( b, r, 1 );
    Fatal( t );
  END;
  ra := a.sin_addr.s_addr;
  rp := WinSock.ntohs( a.sin_port );
  RETURN CARDINAL( r );
END RdSock;


PROCEDURE IncCnt( VAR x: CARDINAL );
BEGIN
  IF x < MAX( CARDINAL ) THEN INC( x ) ELSE x := 0 END;
END IncCnt;


PROCEDURE Install( h: THandler );
BEGIN
  IF hnd # NULLPROC THEN Fatal( 'Install Called twice' ); END;
  hnd := h;
END Install;

PROCEDURE Send( d-: TDstDesc; x-: ARRAY OF BYTE; l: CARDINAL );

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Send ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, ' bytes sent to ' );
    Utils.WrIP( b, d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, d.p, 1 );
    Message( s );
  END Show;

BEGIN
  IF ( l > SIZE( x ) ) OR ( l > lenMax ) THEN Fatal( 'Send Bad length' ) END;
  WrSock( soc, d.a, d.p, x, l );
  IncCnt( cnt.txn );
  Show();
END Send;

PROCEDURE Receive();
VAR
  r: TRxInfDesc;
  x: ARRAY [1..lenMax] OF BYTE;
  l: CARDINAL;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Receive ' );
    SO.WrLngCard( b, l, 1 );
    SO.WrStr( b, ' bytes received from ' );
    Utils.WrIP( b, r.d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, r.d.p, 1 );
    Message( s );
  END Show;

BEGIN
  l := RdSock( soc, r.d.a, r.d.p, x );
  IF hnd # NULLPROC THEN
    IncCnt( cnt.rxn );
    hnd( r, x, l );
  ELSE
    IncCnt( cnt.rxe );
  END;
  Show();
END Receive;

PROCEDURE Execute();
CONST
  mega = 1000 * 1000;
  to = WinSock.TIMEVAL{ period DIV mega, period MOD mega };
VAR
  s: TSockSet;
  r: INTEGER;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  LOOP
    Trigger();
    WinSock.FD_ZERO( s );
    WinSock.FD_SET( soc, s );
    r := WinSock.select( 0, s, NIL, NIL, ADR(to) );  (* readability *)
    IF r = WinSock.SOCKET_ERROR THEN
      SO.Init( b, t );
      SO.WrStr( b, 'Execute select() failed, err ' );
      SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
      Fatal( t );
    ELSIF r > 0 THEN
      (* r = count of ready sockets *)
      Receive();
    ELSE
      (* timeout *)
    END;
  END;
  SO.Init( b, t );
  SO.WrStr( b, 'Execute RETURN' );
  Fatal( t );
END Execute;


VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'DoneModule txn ' );
  SO.WrLngCard( b, cnt.txn, 1 );
  SO.WrStr( b, ' txe ' );
  SO.WrLngCard( b, cnt.txe, 1 );
  SO.WrStr( b, ' rxn ' );
  SO.WrLngCard( b, cnt.rxn, 1 );
  SO.WrStr( b, ' rxe ' );
  SO.WrLngCard( b, cnt.rxe, 1 );
  Message( s );
  oldTerm();
END DoneModule;


PROCEDURE InitWSA();
VAR
  w: WinSock.WSADATA;
  e: INTEGER;
  b: SO.TBuffer;
  s: ARRAY [0..63] OF CHAR;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'InitWSA "' );
    SO.WrStr( b, w.szDescription );
    SO.WrStr( b, '", ver ' );
    b.c.p := '0';
    SO.WrHex( b, w.wVersion, 4 );
    Message( s );
  END Show;

BEGIN
  e := WinSock.WSAStartup( 0202H, w );
  IF e # 0 THEN
    SO.Init( b, s );
    SO.WrStr( b, 'InitWSA WSAStartup() failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( s );
  END;
  Show();
END InitWSA;

PROCEDURE InitSock();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..63] OF CHAR;
  d: TDstDesc;
BEGIN
  soc := NewSock();
  AdjSock( soc );
  BindSock( soc, Config.data.adr, Config.data.prt );
  GetSockAdr( soc, d.a, d.p );
  IF Config.data.vrb THEN
    SO.Init( b, s );
    SO.WrStr( b, 'InitSock Listening on ' );
    Utils.WrIP( b, d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, d.p, 1 );
    Message( s );
  END;
END InitSock;

BEGIN
  Log.Register( mid, 'CliIO' );
  hnd := NULLPROC;
  soc := socNull;
  Lib2.Fill( cnt, 0 );
  tcs := Tmr.Ticks();

  Lib3.Terminate( DoneModule, oldTerm );

  InitWSA();
  InitSock();
END CliIO.
